/*

                                   _JNJ`                        
                               .JNMH`                           
                              JMMF`       `;.                   
                            .NMM)          `MN.                 
                            MMM)            (MML                
                           (MMM`             MMML               
                     M  I  D  N  I  G  H  T    C  o  D  E       
                           (NMMF             MHNH               
                            NMML            .MMM                
                             NMML          .NMH                 
                              4MMNL       .#F                   
                               `4HNNL_   `                      
                                   `"""`                        

                          Copyright (C) 2004-2012
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/


/*
  Hardware;
  
    In Australia, JayCar stocks all of the needed components;
      9v Supply:   http://jaycar.com.au/productView.asp?ID=MP3146
      Patch lead:  http://jaycar.com.au/productView.asp?ID=YN8282
      Etherten:    http://jaycar.com.au/productView.asp?ID=XC4216
           Source: http://www.freetronics.com/products/etherten
      LCD Shield:  http://jaycar.com.au/productView.asp?ID=XC4218
           Source: http://www.freetronics.com/products/lcd-keypad-shield
      FogHorn Kit: http://jaycar.com.au/productView.asp?ID=KG9092
      Speaker:     http://jaycar.com.au/productView.asp?ID=AS3000


  Software;
  
    Compiled via Arduino IDE
      arduino-022: http://arduino.cc/en/Main/Software
      
    With SD, this is a big sketch, so it errors on the Uno boot loader;
      Sketch size 29884: http://arduino.cc/forum/index.php?topic=73207.0
      
    Requires the fixed Optiboot boot loader;
      Optiboot:    http://arduino.cc/forum/index.php/topic,64105.msg466703.html#msg466703
           Source: https://github.com/WestfW/Arduino
                   https://raw.github.com/WestfW/Arduino/master/hardware/arduino/bootloaders/optiboot/optiboot_atmega328.hex
                   https://raw.github.com/WestfW/Arduino/master/hardware/arduino/bootloaders/optiboot/optiboot_atmega328.lst
                
    Which can be installed from another Arduino, thanks to Optifix;
      Optifix:     http://www.adafruit.com/forums/viewtopic.php?f=25&t=19260
          Source:  https://docs.google.com/uc?id=0B6dMB5dovDUZM2ZhMmUxYTItM2ExNy00MWM1LWIzMTgtNWI0ZGUwYTNhZTE2&export=download&hl=en
      ArduinoISP:  http://arduino.cc/en/Tutorial/ArduinoISP
    
    Requires the following add-on/non-standard libraries;
      Ethernet:    http://gkaindl.com/software/arduino-ethernet/dhcp
        NTP patch: http://midnightcode.org/projects/clock/
      Time/Alarm:  http://arduino.cc/playground/Code/Time
      SDFAT:       http://code.google.com/p/sdfatlib/downloads/list

    Thanks also
      LadyAda:     http://www.ladyada.net/learn/arduino/ethfiles.html
    
    Todo;
      Buttons - allow user to set TZ and Alarm data from the LCD side.

    Note that the conflict between SD, Ethernet and LCD hasn't yet 
    been resolved, I currently blame pin 4 conflicts.  Enabling SD 
    code in this sketch will result in LCD and ethernet issues.  I've
    run out of time to track the cause, if you find a cure then 
    please email me and let me know.

    
  Network;
  
    I have DHCP, DNS and NTP served from my LAN gateway. The DHCP 
    option code for NTP in the ISC DHCP daemon is "ntp-servers".
    
*/
#if defined(ARDUINO) && ARDUINO > 18
#include <SPI.h>
#endif
// #include <SdFat.h>
// #include <SdFatUtil.h>
#include <Ethernet.h>
#include <EthernetDHCP.h>
#include <Udp.h>
#include <Time.h>
#include <TimeAlarms.h>
#include <LiquidCrystal.h>


// Custom Variables - your customisation should go here

// Features on/off
int ip_address_is_static  = false;         // true/false - do you want static addressing?
int web_server_is_enabled = true;          // true/false - do you want a web serving clock?

// Display tweaking
String output_time_colon  = ":";
String output_date_colon  = "/";

// Time tweaking
const long timeZoneOffset = 36000L;        // time zone offset, in seconds
int ntpSyncInterval       = 3600;          // frequency of NTP updates, in seconds

// Layer-4 addressing
unsigned int ntpPort      = 123;
unsigned int httpPort     = 80;

// Layer-3 addressing
byte static_ipAddr[]      = { 192, 168, 1, 18 };
byte static_gatewayAddr[] = { 192, 168, 1, 1 };
byte static_dnsAddr[]     = { 192, 168, 1, 1 };
byte static_ntpAddr[]     = { 192, 168, 1, 1 };

// Layer-2 addressing
byte mac[]                = { 0xC0, 0xDE, 0x00, 0x00, 0x00, 0x01 };




// Static Variables
// DHCP
const char* ip_to_str(const uint8_t*);
const byte* ipAddr;
const byte* gatewayAddr;
const byte* dnsAddr;
const byte* ntpAddr;

// NTP
const int NTP_PACKET_SIZE = 48;            // NTP time stamp is in the first 48 bytes
byte ntpBuffer[NTP_PACKET_SIZE];

// Time
time_t prevDisplay = 0;                    // when the digital clock was displayed
unsigned long time_present;

// HTTP
Server http_server(httpPort);
String http_default_file = "index.htm";
String http_default_content_type = "text/plain";
char * http_mime_types[]  = {".htm", "text/html"
                             ".js",  "application/javascript", 
                             ".png", "image/png", 
                            };
#define http_mime_types_count (sizeof(http_mime_types)/sizeof(char *))
String http_output_time;

// LCD
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);

// Alarm
int alarm_pin = 11;
int alarm_cycle_start = 1000;
int alarm_cycle_stop = 50;
int alarm_cycle_shift = 25;
int alarm_cycle_delay = 5000;

// SD
// Sd2Card sd_card;
// SdVolume sd_volume;
// SdFile sd_root;
int sd_ss_pin = 10;
int sd_cs_pin = 4;
int sd_available = 0;



void setup() {
  String output_note;
  
  Serial.begin(9600);
  
  lcd.begin(16, 2);
  lcd.setCursor(0, 0);

  sd_available = 0;
  /*
  lcd_notice(0, "Initialise SD ..");
  pinMode(sd_ss_pin, OUTPUT);
  digitalWrite(sd_ss_pin, HIGH);
  if(!sd_card.init(SPI_HALF_SPEED, sd_cs_pin)) {
    lcd_notice(1, "SD init fail");
    delay(2000);
  } else {
    // initialize a FAT volume
    if(!sd_volume.init(&sd_card)) {
      lcd_notice(1, "Volume fail");
      delay(2000);
    } else {
      // open the root directory^M
      if(!sd_root.openRoot(&sd_volume)) {
        lcd_notice(1, "Mount fail");
        delay(2000);
      } else {
        sd_available = 1;
      }
    }
  }
  */

  if(ip_address_is_static) {
    lcd_notice(0, "Configure IP ..");
  } else {
    lcd_notice(0, "Obtaining IP ..");
  }
  output_note = "MAC:";
  output_note += mac_to_str(mac);
  lcd_notice(1, output_note);

  if(ip_address_is_static) {
    Ethernet.begin(mac, static_ipAddr);
    ipAddr = static_ipAddr;
    gatewayAddr = static_gatewayAddr;
    dnsAddr = static_dnsAddr;
    ntpAddr = static_ntpAddr;
  } else {
    EthernetDHCP.begin(mac);
    ipAddr = EthernetDHCP.ipAddress();
    gatewayAddr = EthernetDHCP.gatewayIpAddress();
    dnsAddr = EthernetDHCP.dnsIpAddress();
    ntpAddr = EthernetDHCP.ntpIpAddress();
  }
  
  output_note = "IP: ";
  output_note += ip_to_str(ipAddr);
  lcd_notice(0, output_note);  
  output_note = "NTP:";
  output_note += ip_to_str(ntpAddr);
  lcd_notice(1, output_note);  

  Udp.begin(ntpPort);
  setSyncProvider(ntp_client);
  while(timeStatus() == timeNotSet)   
     ; // wait until the time is set by the sync provider
  setSyncInterval(ntpSyncInterval);
  
  pinMode(alarm_pin, OUTPUT);     
  Alarm.alarmRepeat(dowMonday,6,45,00,usrAlarm);
  Alarm.alarmRepeat(dowTuesday,6,45,00,usrAlarm);
  Alarm.alarmRepeat(dowWednesday,6,45,00,usrAlarm);
  Alarm.alarmRepeat(dowThursday,6,45,00,usrAlarm);
  Alarm.alarmRepeat(dowFriday,6,45,00,usrAlarm);
  // Alarm.alarmRepeat(dowSaturday,11,17,00,usrAlarm);

  if(web_server_is_enabled) {
    http_server.begin();
  }
  
  lcd_blank();
}



void alarm_cycle(int on_delay, int off_delay) {
  digitalWrite(alarm_pin, HIGH);
  delay(on_delay);
  digitalWrite(alarm_pin, LOW);
  delay(off_delay);
  return;
}



void alarm_sounder() {
  int alarm_loop;
  for (alarm_loop = alarm_cycle_start; 
    alarm_loop > alarm_cycle_stop; 
    alarm_loop -= alarm_cycle_shift)  {
    alarm_cycle(alarm_cycle_shift, alarm_loop);
  }
  for (alarm_loop = alarm_cycle_start; 
    alarm_loop > alarm_cycle_stop; 
    alarm_loop -= alarm_cycle_shift)  {
    alarm_cycle(alarm_loop, alarm_loop);
  }
  for (alarm_loop = alarm_cycle_start; 
    alarm_loop > alarm_cycle_stop; 
    alarm_loop -= 20)  {
    alarm_cycle(alarm_cycle_stop, alarm_cycle_stop);
  }
  alarm_cycle(alarm_cycle_start * 2, alarm_cycle_start);
  delay(alarm_cycle_delay);
}



void usrAlarm() {
  // lcd_blank();
  lcd_notice(1, "***** ALARM *****");
  alarm_sounder();
}



void loop() {
  if(!ip_address_is_static) {
    EthernetDHCP.maintain();
  }
  if(web_server_is_enabled) {
    web_server();  
  }
  lcd_update();
  Alarm.delay(500);
}



void lcd_blank() {
  String output_note = "                ";
  lcd_notice(0, output_note);
  lcd_notice(1, output_note);
}



void lcd_notice(int row, String message) {
  Serial.print("Notice: ");
  Serial.println(message);
  lcd.setCursor(0, row);
  lcd.print(message);
}



void lcd_update() {
  time_present = now();
  if(time_present != prevDisplay) {
    prevDisplay = time_present;
    digitalClockDisplay();  
  }
} 



void digitalClockDisplay() {
  String output_time_hour;
  String output_time_minute;
  String output_time_second;
  String output_time_ampm;
  String output_time_clock;
  String output_time_year;
  String output_time_month;
  String output_time_day;
  String output_time_weekday;
  String output_time_date;
  String output_time;
  
  output_time_hour = hourFormat12();
  if(output_time_hour.length() == 1) {
    output_time_hour = " " + output_time_hour;
  }
  output_time_minute = minute();
  if(output_time_minute.length() == 1) {
    output_time_minute = "0" + output_time_minute;
  }
  output_time_second = second();
  if(output_time_second.length() == 1) {
    output_time_second = "0" + output_time_second;
  }
  if(isAM()) {
    output_time_ampm = "a";
  } else {
    output_time_ampm = "p";
  }
  output_time_year = year();
  output_time_month = month();
  if(output_time_month.length() == 1) {
    output_time_month = "0" + output_time_month;
  }
  output_time_day = day();
  if(output_time_day.length() == 1) {
    output_time_day = "0" + output_time_day;
  }
  switch(weekday()) {
    case 1:
      output_time_weekday = "Sun";
      break;
    case 2:
      output_time_weekday = "Mon";
      break;
    case 3:
      output_time_weekday = "Tue";
      break;
    case 4:
      output_time_weekday = "Wed";
      break;
    case 5:
      output_time_weekday = "Thu";
      break;
    case 6:
      output_time_weekday = "Fri";
      break;
    case 7:
      output_time_weekday = "Sat";
      break;
  }
  output_time_clock = output_time_hour + output_time_colon + output_time_minute; // + output_time_colon + output_time_second;
  output_time_date = output_time_day + output_date_colon + output_time_month;
  output_time = output_time_clock + output_time_ampm + " " + output_time_weekday + " " + output_time_date;
  
  Serial.println(output_time);
  lcd.setCursor(0, 0);
  lcd.print(output_time);
  
  http_output_time = output_time;
  
  return;
}



unsigned long ntp_client() {
  const unsigned long seventyYears = 2208988800UL;
  unsigned long epoch = 0;
  unsigned long epoch_tz_corrected = 0;

  byte timeServer[] = {  0, 0, 0, 0 };
  timeServer[0] = ntpAddr[0];
  timeServer[1] = ntpAddr[1];
  timeServer[2] = ntpAddr[2];
  timeServer[3] = ntpAddr[3];

  // Send NTP request
  sendNTPpacket(timeServer);
  delay(1000);  
  if ( Udp.available() ) {  
    Udp.readPacket(ntpBuffer, NTP_PACKET_SIZE);
    // the timestamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, esxtract the two words:
    unsigned long highWord = word(ntpBuffer[40], ntpBuffer[41]);
    unsigned long lowWord = word(ntpBuffer[42], ntpBuffer[43]);  
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900):
    unsigned long secsSince1900 = highWord << 16 | lowWord;  
    Serial.print("Seconds since Jan 1 1900 = " );
    Serial.println(secsSince1900);               

    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
    // subtract seventy years:
    epoch = secsSince1900 - seventyYears;
    epoch_tz_corrected = epoch + timeZoneOffset;  

    // print the hour, minute and second:
    Serial.print("The UTC time is ");       // UTC is the time at Greenwich Meridian (GMT)
    Serial.print((epoch  % 86400L) / 3600); // print the hour (86400 equals secs per day)
    Serial.print(':');  
    Serial.print((epoch  % 3600) / 60);     // print the minute (3600 equals secs per minute)
    Serial.print(':'); 
    Serial.println(epoch %60);              // print the second
    Serial.print("The LOCAL time is ");
    Serial.print((epoch_tz_corrected  % 86400L) / 3600);
    Serial.print(':');  
    Serial.print((epoch_tz_corrected  % 3600) / 60);
    Serial.print(':'); 
    Serial.println(epoch_tz_corrected %60);

  }
  Serial.println("-------------------------------");
  
  return epoch_tz_corrected;
}



void web_server_file(Client client, String file_name) {
  unsigned int error_state;
  String error_message;
  char sd_file_name[12]; 
  unsigned int mime_index;
  String content_type;
  String output_content_type;
  // SdFile file;
  int16_t file_byte;
  
  Serial.println("Web request: " + file_name);
  
  // Basic web server content until SD issue is resolved
  client.println("HTTP/1.1 200 OK");
  client.println("Content-Type: text/html");
  client.println();
  client.println("Time: " + http_output_time);
    
/*
  error_state = 1;
  if(!sd_available) {
    error_message = "No SD capability";
  } else {
    if(file_name.length() < 1 || file_name.length() > 12) {
      error_message = "Invalid file name";
    } else {
      sprintf(sd_file_name, "%s", file_name);
      if(!file.open(&sd_root, sd_file_name, O_READ)) {
        error_message = "Unable to open file";
      } else {
        error_state = 0;
      }
    }
  }
  if(error_state) {
    client.println("HTTP/1.1 404 Not Found");
    client.println("Content-Type: text/html");
    client.println();
    client.println("<h2>File not found .. </h2>");
    client.println("<br><br>");
    client.println("&nbsp; &nbsp; <h3>" + error_message + "</h3>");
    return;
  }

  // Find content type for extension
  content_type = http_default_content_type;
  for(mime_index = 0; mime_index < http_mime_types_count / 2; mime_index += 2) {
    if(file_name.lastIndexOf(http_mime_types[mime_index]) != -1) {
      content_type = http_mime_types[mime_index + 1];
      break;
    }
  }
  output_content_type = "Content-Type: " + content_type;

  // Run output sequence
  client.println("HTTP/1.1 200 OK");
  client.println(output_content_type);
  client.println();
  while ((file_byte = file.read()) > 0) {
    // Serial.print((char)c);
    if(file_name == "index.htm" && file_byte == "@") {
      client.print(http_output_time);
    } else {
      client.print((char)file_byte);
    }
  }
  file.close();
  */
          
  return;
}

  
void web_server() {
  boolean client_input_empty = true;
  unsigned int client_input_max = 128;
  char client_input[client_input_max];
  int client_input_index = 0;
  String filename;
  
  // listen for incoming clients
  Client client = http_server.available();
  if (client) {
    client_input_empty = true;
    while (client.connected()) {
      if (client.available()) {
        Serial.println("Web connection ...");
        char c = client.read();
        if (c != '\n' && c != '\r') {
          client_input[client_input_index] = c;
          client_input_index++;
          if (client_input_index >= client_input_max) 
            client_input_index = client_input_max - 1;
          continue;
        }
        client_input[client_input_index] = 0;
        // Look for file request substring
        if (strstr(client_input, "GET / ") != 0) {
          web_server_file(client, http_default_file);
          
        } else if (strstr(client_input, "GET /") != 0) {
          // this time no space after the /, so a sub-file!
          (strstr(client_input, " HTTP"))[0] = 0; // replace " HTTP/1.1" with \0
          filename = String(client_input + 5);    // ignore the "GET /" (5 chars)
          web_server_file(client, filename);
 
        } else {
          web_server_file(client, "404_ERR");

        }
        break;
      }
    }
    delay(5);
    client.stop();
  }
}




const char* ip_to_str(const uint8_t* ipAddr) {
  static char buf[16];
  sprintf(buf, "%d.%d.%d.%d\0", ipAddr[0], ipAddr[1], ipAddr[2], ipAddr[3]);
  return buf;
}



const char* mac_to_str(const uint8_t* macAddr) {
  static char buf[16];
  sprintf(buf, "%02x%02x%02x%02x%02x%02x\0", 
    macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]);
  return buf;
}



unsigned long sendNTPpacket(byte *address) {
  // Build NTP request
  memset(ntpBuffer, 0, NTP_PACKET_SIZE); 
  ntpBuffer[0]   = 0b11100011; // LI, Version, Mode
  ntpBuffer[1]   = 0;          // Stratum, or type of clock
  ntpBuffer[2]   = 6;          // Polling Interval
  ntpBuffer[3]   = 0xEC;       // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  ntpBuffer[12]  = 49; 
  ntpBuffer[13]  = 0x4E;
  ntpBuffer[14]  = 49;
  ntpBuffer[15]  = 52;
  Udp.sendPacket(ntpBuffer, NTP_PACKET_SIZE, address, 123);
}



